# IEDMod release 1 by David Bowland
# ./addons/eventscripts/iedmod/iedmod.py

# >>> To configure this addon please see iedmod.cfg <<<

"""
Allows players to create props which explode when players are near. Requires only ES 2.0+
"""


import es
import gamethread
import langlib
import os.path
import playerlib
import random


info = es.AddonInfo()
info.name = 'IEDMod'
info.version = '1'
info.url = 'http://addons.eventscripts.com/addons/view/iedmod'
info.basename = 'iedmod'
info.author = 'SuperDave'


list_ied_models = []
list_file_downloads = []
dict_say_commands = {'purchase':'', 'create':''}
dict_players = {}
dict_ieds = {}

dict_options = {'iedmod_detonation_radius':[150, 'IEDs will be detonated when a player is within this distance'], 'iedmod_detonation_checkdelay':[1, 'Number of seconds between checking the distance of players'], 'iedmod_detonation_team':[0, '0 = Only opposing team members can detonate an IED, 1 = Any player can detonate an IED'], 'iedmod_explosion_radius':[350, 'Radius of IED explosion--HE grenade is 350'], 'iedmod_explosion_magnitude':[200, 'Magnitude of IED explosion--HE grenade is 100'], 'iedmod_explosion_sound':[40, 'Percent of the map that can hear an IED explode'], 'iedmod_purchase_cost':[2000, 'Cost of purchasing an IED--0 will give all players the maximum number of IEDs'], 'iedmod_purchase_max':[3, 'Maximum number of IEDs a player can have--0 for no limit'], 'iedmod_purchase_command':['!buyied', 'Command players must say to purchase an IED'], 'iedmod_purchase_buyzone':[1, '0 = IEDs can be puchased anywhere, 1 = IEDs can only be puchased in buy zones'], 'iedmod_purchase_loseafterdeath':[1, '0 = players keep IEDs that have not been planted when killed, 1 = players will lose IEDs that have not been planted when killed'], 'iedmod_create_command':['!ied', 'Command players must say to plant an IED'], 'iedmod_create_delay':[5, 'Number of seconds a player must wait after creating an IED to create another--set to 0 for no delay'], 'iedmod_create_knife':[1, '0 = players can create an IED with any active weapon, 1 = players are required to have the knife as active weapon to create an IED, 2 = players will be forced to have the knife as active weapon when creating and IED [default 1]'], 'iedmod_bot_allowied':[2, 'Number of IEDs bot will be allowed to buy or plant per round'], 'iedmod_bot_purchasechance':[80, 'Percent chance a bot will purchase an IED if it has enough money'], 'iedmod_bot_createdelay_min':[5, 'Minumim number of seconds after spawning a bot will wait to create an IED'], 'iedmod_bot_createdelay_max':[45, 'Maximum number of seconds after spawning a bot will wait to create an IED'], 'iedmod_bot_iterationstoremove':[4, 'Number of IED distance updates until an IED spawned by a bot is removed if that bot is stuck in the IED'], 'iedmod_announce_spawn':[1, '0 = no change, 1 = announce to each player his or her IED count as well as how to purchase and create IEDs when the player spawns'], 'iedmod_announce_purchase':[2, '0 = no change, 1 = announce IED purchasing feedback in chat area, 2 = announce IED purchasing feedback in center text'], 'iedmod_announce_create':[2, '0 = no change, 1 = announce IED creation feedback in chat area, 2 = announce IED creation feedback in center text']}
func_lang_text = None


def load():
   """
   Creates commands for settings
   Checks to make sure critical server variables are created by the config
   Creates say commands for purchasing and creating IEDs
   Calls es_map_start
   Calls player_spawn for each live player
   """
   global dict_say_commands
   global func_lang_text

   if not es.exists('command', 'iedmod_add'):
      es.regcmd('iedmod_add', 'iedmod/model_cmd', 'iedmod_add <model relative to ./models/>\nServer command for adding IED models')
   if not es.exists('command', 'iedmod_remove'):
      es.regcmd('iedmod_remove', 'iedmod/model_cmd', 'iedmod_remove <model relative to ./models/>\nServer command for removing IED models')
   if not es.exists('command', 'iedmod_download_add'):
      es.regcmd('iedmod_download_add', 'iedmod/download_cmd', 'iedmod_download_add <file relative to game directory>\nServer command for adding IED files for client download')
   if not es.exists('command', 'iedmod_download_remove'):
      es.regcmd('iedmod_download_remove', 'iedmod/download_cmd', 'iedmod_download_remove <file relative to game directory>\nServer command for removing IED files for client download')

   for str_option in dict_options:
      es.ServerVar(str_option, dict_options[str_option][0], dict_options[str_option][1])
   if os.path.isfile(es.getAddonPath('iedmod') + '/iedmod.cfg'):
      es.server.cmd('es_xmexec ../addons/eventscripts/iedmod/iedmod.cfg')
   else:
      es.dbgmsg(0, 'IEDMod: Unable to load iedmod.cfg! Please ensure it is in the ./iedmod/ directory.')
   str_lang_path = es.getAddonPath('iedmod') + '/iedmod_languages.ini'
   if os.path.isfile(str_lang_path):
      func_lang_text = langlib.Strings(str_lang_path)
   else:
      es.dbgmsg(0, 'IEDMod: Unable to load iedmod_languages.ini! Please ensure it is in the ./iedmod/ directory.')
   if not func_lang_text:
      func_lang_text = missing_lang

   str_command = str(es.ServerVar('iedmod_purchase_command'))
   if str_command:
      if not es.exists('saycommand', str_command):
         es.regsaycmd(str_command, 'iedmod/buy_ied_cmd', 'Say command for purchasing an IED')
         dict_say_commands['purchase'] = str_command
   str_command = str(es.ServerVar('iedmod_create_command'))
   if str_command:
      if not es.exists('saycommand', str_command):
         es.regsaycmd(str_command, 'iedmod/create_ied_cmd', 'Say command for creating an IED')
         dict_say_commands['create'] = str_command

   es_map_start({})

   for int_userid in es.getUseridList():
      if not es.getplayerprop(int_userid, 'CBasePlayer.pl.deadflag'):
         player_spawn({'userid':str(int_userid), 'es_steamid':es.getplayersteamid(int_userid), 'es_userteam':str(es.getplayerteam(int_userid))})


def es_map_start(event_var):
   """
   Cancels outstanding delays from the previous map
   Clears the dictionary of player
   Adds files that need to be downloaded by clients to the "downloadables" string table
   """
   global dict_players

   for int_userid in dict_players:
      if dict_players[int_userid]['prevent_create']:
         gamethread.cancelDelayed('iedmod_player_%s' % int_userid)
   gamethread.cancelDelayed('iedmod_updateloop')

   dict_players.clear()

   for str_file in list_file_downloads:
      es.stringtable('downloadables', str_file)


def round_start(event_var):
   """
   Removes outstanding IEDs and corresponding delays
   Removes outstanding IED creation delays
   Clears the dictionary of IEDs
   """
   global dict_players
   global dict_ieds

   dict_active_entites = es.createentitylist()
   for int_userid in dict_players:
      dict_current_player = dict_players[int_userid]
      list_active_ieds = dict_current_player['active_ieds']
      for int_ied_index in list_active_ieds:
         if dict_active_entites.has_key(int_ied_index):
            es.server.cmd('es_xremove %s' % int_ied_index)
      list_active_ieds[:] = []

      if dict_current_player['prevent_create'] == 1:
         gamethread.cancelDelayed('iedmod_player_%s' % int_userid)
         dict_current_player['prevent_create'] = 0

   dict_ieds.clear()


def player_spawn(event_var):
   """
   Creates the player in the player dictionary
   Calls spawn_announce
   Handles purchase and creation of IEDs by bots
   """
   global dict_players

   int_userid = int(event_var['userid'])
   int_team = int(event_var['es_userteam'])
   if not dict_players.has_key(int_userid):
      dict_players[int_userid] = {'ied_count':0, 'active_ieds':[], 'prevent_create':0}

   dict_current_player = dict_players[int_userid]
   dict_current_player['opposing_team'] = 5 - int_team
   if int_team > 1:
      if not int(es.ServerVar('iedmod_purchase_cost')):
         dict_current_player['ied_count'] = int(es.ServerVar('iedmod_purchase_max'))

      spawn_announce(int_userid)

      if event_var['es_steamid'] == 'BOT' and int(es.ServerVar('iedmod_bot_allowied')):
         if dict_current_player['prevent_create']:
            gamethread.cancelDelayed('iedmod_player_%s' % int_userid)
         int_ied_cost = int(es.ServerVar('iedmod_purchase_cost'))
         if int_ied_cost:
            while random.randint(1, 100) <= es.ServerVar('iedmod_bot_purchasechance') and es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount') >= int_ied_cost + int(es.ServerVar('bot_eco_limit')) and dict_current_player['ied_count'] < int(es.ServerVar('iedmod_bot_allowied')):
               es.sexec(int_userid, 'say %s' % dict_say_commands['purchase'])
         else:
            dict_current_player['ied_count'] = int(es.ServerVar('iedmod_bot_allowied'))
         if dict_current_player['ied_count']:
            dict_current_player['prevent_create'] = -1
            gamethread.delayedname(random.randint(es.ServerVar('iedmod_bot_createdelay_min'), es.ServerVar('iedmod_bot_createdelay_max')), 'iedmod_player_%s' % int_userid, bot_create_ied, int_userid)


def player_death(event_var):
   """
   Removes the dead player's delay
   Removes the dead player's IED count
   """
   global dict_players

   int_userid = int(event_var['userid'])
   if dict_players.has_key(int_userid):
      dict_current_player = dict_players[int_userid]
      if dict_current_player['prevent_create']:
         gamethread.cancelDelayed('iedmod_player_%s' % int_userid)
         dict_current_player['prevent_create'] = 0

      if int(es.ServerVar('iedmod_purchase_loseafterdeath')):
         dict_current_player['ied_count'] = 0


def player_disconnect(event_var):
   """Removes the disconneting player's IEDs and delays"""
   global dict_players
   global dict_ieds

   int_userid = int(event_var['userid'])
   if dict_players.has_key(int_userid):
      list_active_ieds = dict_players[int_userid]['active_ieds']
      for int_ied_index in list_active_ieds:
         gamethread.cancelDelayed('iedmod_ied_%s' % int_ied_index)
         es.server.cmd('es_xremove %s' % int_ied_index)
         del dict_ieds[int_ied_index]
      if dict_players[int_userid]['prevent_create']:
         gamethread.cancelDelayed('iedmod_player_%s' % int_userid)
      del dict_players[int_userid]


def unload():
   """
   Removes active IEDs and delays
   Cancels ied update delay
   Unregisters say commands
   """
   dict_active_entites = es.createentitylist()
   for int_userid in dict_players:
      dict_current_player = dict_players[int_userid]
      list_active_ieds = dict_current_player['active_ieds']
      for int_ied_index in list_active_ieds:
         if dict_active_entites.has_key(int_ied_index):
            es.server.cmd('es_xremove %s' % int_ied_index)
      if dict_current_player['prevent_create']:
         gamethread.cancelDelayed('iedmod_player_%s' % int_userid)

   if dict_ieds:
      gamethread.cancelDelayed('iedmod_updateloop')

   for str_command_type in dict_say_commands:
      es.unregsaycmd(dict_say_commands[str_command_type])

   for str_option in dict_options:
      es.ServerVar(str_option).set(0)


def buy_ied_cmd():
   """
   Increases the player's IED count
   Displays IED purchase status
   """
   global dict_players

   int_userid = es.getcmduserid()
   int_announce_option = int(es.ServerVar('iedmod_announce_purchase'))
   if es.getplayerteam(int_userid) > 1 and not es.getplayerprop(int_userid, 'CBasePlayer.pl.deadflag'):
      if es.getplayerprop(int_userid, 'CCSPlayer.m_bInBuyZone') or not int(es.ServerVar('iedmod_purchase_buyzone')):
         int_ied_max = int(es.ServerVar('iedmod_purchase_max'))
         dict_current_player = dict_players[int_userid]
         if dict_current_player['ied_count'] < int_ied_max or not int_ied_max:
            int_money = es.getplayerprop(int_userid, 'CCSPlayer.m_iAccount')
            int_cost = int(es.ServerVar('iedmod_purchase_cost'))
            if int_cost <= int_money and int_cost:
               es.setplayerprop(int_userid, 'CCSPlayer.m_iAccount', int_money - int_cost)
               dict_current_player['ied_count'] += 1

               if int_announce_option:
                  announce_type(int_announce_option, int_userid, 'purchase success', {})
                  spawn_announce(int_userid)
            else:
               announce_type(int_announce_option, int_userid, 'purchase cash', {'cost':int_cost})
         else:
            announce_type(int_announce_option, int_userid, 'purchase max', {})
      else:
         announce_type(int_announce_option, int_userid, 'purchase buy zone', {})
   else:
      announce_type(int_announce_option, int_userid, 'purchase dead', {})


def create_ied_cmd():
   """
   Checks the player for an IED to create
   Creates the IED and starts the IED loop
   Starts the delay to prevent players from creating a second IED too quickly
   Announces IED creation status
   """
   global dict_players
   global dict_ieds

   int_userid = es.getcmduserid()
   int_announce_option = int(es.ServerVar('iedmod_announce_create'))
   if es.getplayerteam(int_userid) > 1 and not es.getplayerprop(int_userid, 'CBasePlayer.pl.deadflag'):
      dict_current_player = dict_players[int_userid]
      if not dict_current_player['prevent_create']:
         int_ied_max = int(es.ServerVar('iedmod_purchase_max'))
         if dict_current_player['ied_count'] or not int_ied_max:
            int_require_knife = int(es.ServerVar('iedmod_create_knife'))
            if int_require_knife == 2 or es.getplayersteamid(int_userid) == 'BOT':
                  es.sexec(int_userid, 'use weapon_knife')
                  int_require_knife = 0
            if es.createplayerlist(int_userid)[int_userid]['weapon'] == 'weapon_knife' or not int_require_knife:
               if dict_current_player['ied_count']:
                  dict_current_player['ied_count'] -= 1

               es.server.cmd('es_xprop_physics_create %s %s' % (int_userid, random.choice(list_ied_models)))
               int_ied_index = int(es.ServerVar('eventscripts_lastgive'))
               dict_current_player['active_ieds'].append(int_ied_index)
               dict_ieds[int_ied_index] = {'owner':int_userid, 'prevent_explosion':1}
               if len(dict_ieds) == 1:
                  gamethread.delayedname(es.ServerVar('iedmod_detonation_checkdelay'), 'iedmod_updateloop', ied_loop)

               int_creation_delay = int(es.ServerVar('iedmod_create_delay'))
               if int_creation_delay:
                  dict_current_player['prevent_create'] = 1
                  gamethread.delayedname(int_creation_delay, 'iedmod_player_%s' % int_userid, remove_restriction, int_userid)

               announce_type(int_announce_option, int_userid, 'create success', {})
               if int_announce_option and int(es.ServerVar('iedmod_announce_spawn')):
                  int_max = int(es.ServerVar('iedmod_purchase_max'))
                  if int_max:
                     es.tell(int_userid, '#multi', func_lang_text('ied count', {'count':dict_current_player['ied_count'], 'max':int_max}, playerlib.getPlayer(int_userid).get('lang')))
                  else:
                     es.tell(int_userid, '#multi', func_lang_text('ied count no max', {'count':dict_current_player['ied_count']}, playerlib.getPlayer(int_userid).get('lang')))
            else:
               announce_type(int_announce_option, int_userid, 'create knife', {})
         else:
            announce_type(int_announce_option, int_userid, 'create empty', {})
      else:
         announce_type(int_announce_option, int_userid, 'create limit', {})
   else:
      announce_type(int_announce_option, int_userid, 'create dead', {})


def ied_loop():
   """
   Checks the IED for players inside the detonation radius
   Detonates the IED and removes the prop
   """
   global dict_players
   global dict_ieds

   list_ieds_to_remove = []
   dict_prop_physics = es.createentitylist('prop_physics')
   dict_player_list = es.createplayerlist()
   for int_ied_index in dict_ieds:
      dict_current_ied = dict_ieds[int_ied_index]
      int_ied_owner = dict_current_ied['owner']
      dict_current_player = dict_players[int_ied_owner]
      if dict_prop_physics.has_key(int_ied_index):
         str_ied_origin = es.getindexprop(int_ied_index, 'CBaseEntity.m_vecOrigin')
         list_ied_location = str_ied_origin.split(',')
         if dict_current_ied['prevent_explosion']:
            tuple_player_location = es.getplayerlocation(int_ied_owner)
            if pow(pow(float(list_ied_location[0]) - tuple_player_location[0], 2) + pow(float(list_ied_location[1]) - tuple_player_location[1], 2) + pow(float(list_ied_location[2]) - tuple_player_location[2], 2), 0.5) < es.ServerVar('iedmod_detonation_radius'):
               dict_current_ied['prevent_explosion'] += 1
               if es.getplayersteamid(int_ied_owner) == 'BOT' and dict_current_ied['prevent_explosion'] >= es.ServerVar('iedmod_bot_iterationstoremove'):
                  dict_current_player['active_ieds'].remove(int_ied_index)
                  list_ieds_to_remove.append(int_ied_index)
                  es.server.cmd('es_xremove %s' % int_ied_index)
               continue
            else:
               dict_current_ied['prevent_explosion'] = 0

         for int_userid in dict_player_list:
            dict_loop_player = dict_player_list[int_userid]
            if (dict_loop_player['teamid'] == dict_current_player['opposing_team'] or (int(es.ServerVar('iedmod_detonation_team')) and dict_loop_player['teamid'] in (2, 3))) and dict_loop_player['isdead'] == 0:
               if pow(pow(float(list_ied_location[0]) - dict_loop_player['x'], 2) + pow(float(list_ied_location[1]) - dict_loop_player['y'], 2) + pow(float(list_ied_location[2]) - dict_loop_player['z'], 2), 0.5) < es.ServerVar('iedmod_detonation_radius'):
                  for int_explosion_index in es.createentitylist('env_explosion'):
                     es.server.cmd('es_xremove %s' % int_explosion_index)
                  es.emitsound('entity', int_ied_index, 'weapons/hegrenade/explode%s.wav' % random.randint(3, 5), '1.0', (100 - float(es.ServerVar('iedmod_explosion_sound'))) / 100)
                  dict_current_player['active_ieds'].remove(int_ied_index)
                  list_ieds_to_remove.append(int_ied_index)
                  es.server.cmd('es_xremove %s' % int_ied_index)

                  es.server.cmd('es_xgive %s env_explosion' % int_userid)
                  int_explosion_index = int(es.ServerVar('eventscripts_lastgive'))
                  es.server.cmd('es_xfire %s env_explosion addoutput \"imagnitude %s\"' % (int_userid, es.ServerVar('iedmod_explosion_magnitude')))
                  es.server.cmd('es_xfire %s env_explosion addoutput \"iradiusoverride %s\"' % (int_userid, es.ServerVar('iedmod_explosion_radius')))
                  es.setindexprop(int_explosion_index, 'CBaseEntity.m_vecOrigin', str_ied_origin)
                  es.setindexprop(int_explosion_index, 'CBaseEntity.m_hOwnerEntity', es.getplayerhandle(int_ied_owner))
                  es.server.cmd('es_xfire %s env_explosion explode' % int_userid)
                  es.server.cmd('es_xfire %s env_explosion kill' % int_userid)
                  break
      else:
         if int_ied_index in dict_current_player['active_ieds']:
            dict_current_player['active_ieds'].remove(int_ied_index)
         list_ieds_to_remove.append(int_ied_index)
   for int_ied_index in list_ieds_to_remove:
      del dict_ieds[int_ied_index]

   if dict_ieds:
      gamethread.delayedname(es.ServerVar('iedmod_detonation_checkdelay'), 'iedmod_updateloop', ied_loop)


def remove_restriction(int_userid):
   """Removes the restriction for creating another IED"""
   global dict_players

   if dict_players.has_key(int_userid):
      dict_players[int_userid]['prevent_create'] = 0


def bot_create_ied(int_userid):
   """Forces the provided bot to create an IED"""
   global dict_players

   if dict_players.has_key(int_userid):
      dict_current_player = dict_players[int_userid]
      dict_current_player['prevent_create'] = 0

      es.sexec(int_userid, 'say %s' % dict_say_commands['create'])

      if dict_current_player['ied_count']:
         dict_current_player['prevent_create'] = -1
         gamethread.delayedname(random.randint(es.ServerVar('iedmod_bot_createdelay_min'), es.ServerVar('iedmod_bot_createdelay_max')), 'iedmod_player_%s' % int_userid, bot_create_ied, int_userid)


def model_cmd():
   """
   iedmod_add <model relative to ./models/>
   iedmod_remove <model relative to ./models/>
   Adds or removes models from the list of IED models
   """
   global list_ied_models

   str_command = es.getargv(0).lower()
   if es.getargc() == 2:
      str_file = es.getargv(1)
      int_file_length = len(str_file)
      if str_file[int_file_length - 4:int_file_length].lower() <> '.mdl':
         str_file += '.mdl'
      if str_command == 'iedmod_add':
         if str_file not in list_ied_models:
            list_ied_models.append(str_file)
      elif str_file in list_ied_models:
         list_ied_models.remove(str_file)
   else:
      es.dbgmsg(0, 'Syntax: ' + {'iedmod_add':'iedmod_add <model relative to ./models/>', 'iedmod_remove':'iedmod_remove <model relative to ./models/>'}[str_command])


def download_cmd():
   """
   iedmod_download_add <file relative to game directory>
   iedmod_download_remove <file relative to game directory>
   Adds or removes files from the list of file automatically downloaded by clients
   """
   global list_file_downloads

   str_command = es.getargv(0).lower()
   if es.getargc() == 1:
      str_file = es.getargv(1)
      if str_command == 'iedmod_download_add':
         if str_file not in list_file_downloads:
            list_file_downloads.append(str_file)
      elif str_file in list_file_downloads:
         list_file_downloads.remove(str_file)
   else:
      es.dbgmsg(0, 'Syntax: ' + {'iedmod_download_add':'iedmod_download_add <file relative to game directory>', 'iedmod_download_remove':'iedmod_download_remove <file relative to game directory>'}[str_command])


def announce_type(int_announce_option, int_userid, str_text, dict_tokens):
   """Announces the provided text in the desired manner"""
   if int_announce_option:
      if int_announce_option == 2:
         es.centertell(int_userid, remove_tags(func_lang_text(str_text, dict_tokens, playerlib.getPlayer(int_userid).get('lang'))))
      else:
         es.tell(int_userid, '#multi', func_lang_text(str_text, dict_tokens, playerlib.getPlayer(int_userid).get('lang')))


def spawn_announce(int_userid):
   """
   Displays the player's IED count
   Displays instructions on purchasing and creating IEDs
   """
   if int(es.ServerVar('iedmod_announce_spawn')):
      dict_current_player = dict_players[int_userid]
      int_max = int(es.ServerVar('iedmod_purchase_max'))
      if int_max:
         es.tell(int_userid, '#multi', func_lang_text('ied count', {'count':dict_current_player['ied_count'], 'max':int_max}, playerlib.getPlayer(int_userid).get('lang')))
      else:
         es.tell(int_userid, '#multi', func_lang_text('ied count no max', {'count':dict_current_player['ied_count']}, playerlib.getPlayer(int_userid).get('lang')))

      dict_tokens = dict_say_commands.copy()
      dict_tokens['cost'] = int(es.ServerVar('iedmod_purchase_cost'))
      if (dict_current_player['ied_count'] < int_max or not int_max) and dict_tokens['cost']:
         if dict_current_player['ied_count'] or not int_max:
            es.tell(int_userid, '#multi', func_lang_text('ied create or purchase', dict_tokens, playerlib.getPlayer(int_userid).get('lang')))
         else:
            es.tell(int_userid, '#multi', func_lang_text('ied purchase', dict_tokens, playerlib.getPlayer(int_userid).get('lang')))
      else:
         es.tell(int_userid, '#multi', func_lang_text('ied create', dict_tokens, playerlib.getPlayer(int_userid).get('lang')))


def remove_tags(str_text):
   """Removes #lightgreen, #green and #default tags from the supplied string"""
   return str_text.replace('#lightgreen', '').replace('#green', '').replace('#default', '')


def missing_lang(str_heading, dict_tokens, str_language):
   """Displays an error message because the languages .ini could not be found"""
   return 'IEDMod: iedmod_languages.ini not found'